{ TFastFormula -
  Copyright by Frank Kroeger   FKroeger@tkc.de }

{ This file show the records and objects to use with TFastFormula }

{ this record is returned by TFastFormula.Compile
  Code : If <>0 an error has occured.
  Position : where did the error occur
  Text : a small description of what happened}
  TFastFormulaError = record
     Code : Byte;
     Position : LongInt;
     Text : AnsiString;
  end;



{ This list handles user added constants }
{ You can either use the object inspector to add new constants, or
  TfastFormula.Constants.Add ( Name, Value, Formula );
  Name : Name of the new constant
  Value : the value of the constant
  Formula : formula to calculate the value, this is ONLY used by the object inspector
  }
  TConstantsList = class(TPersistent)
   public
     procedure Clear;
     procedure Add(const Name:AnsiString;const Value:Extended;const Formula:AnsiString);
     property Count:LongInt read FCount write FCount;
  end;

{ These are the event handlers used by TFastFormula }

{ TValidateParameter is called when TFastFormula finds a parameter in the formula string }
  TValidateParameter = procedure(Sender:TObject;Parameter:AnsiString;var Valid:Boolean) of object;

{ TGetParameterValue is called to get a value for a given parameter. If you use this event
  you HAVE to use TFastFormula.Calculate3 to receive the call to TGetParameterValue }
  TGetParameterValue = procedure(Sender:TObject;Parameter:AnsiString;var Value:Extended) of object;

type
  TFastFormula = class(TComponent)
  public
    { Public-Deklarationen }
    { Calculate is used if you want to set the parametervalues when you calculate }
    Calculate : function(const A : Array of Extended):Extended;
    { Calculate2 is used when the formula does not have any variables, only constant values }
    Calculate2 : function:Extended;
    { Calculate3 has to be used when the event-handler for TGetParametervalue is used }
    function Calculate3:Extended;
    constructor Create(AOwner:TComponent);override;
    { Compile : This function converts the formula string to real compiled and optimized code }
    function Compile:TFastFormulaError;
{$IFNDEF SHAREWARE}
    procedure ShowUPN;  // This procedure show the reverse polish notation of the formula (only in the registered version
{$ENDIF}
    { This procedure returns the code TFastFormula generated. (in mnemonics) only in registered version }
    procedure GetSourceCode(var StrList:TStringList);
    { Add a new function to TFastFormulas function list. SEE BELOW }
    procedure AddFunction(Name:AnsiString;
                          NumberOfParameters:LongInt;
                          Address:pointer;
                          StackUsage:Byte;
                          SaveStack:Boolean);
    procedure AddInlineFunction(Name:AnsiString;
                                NumberOfParameters:LongInt;
                                StackUsage:Byte;
                                ParametersInStack:Boolean;CodeSize:longInt;
                                Code:pointer;
                                RegisterUsage:Byte);
    { Returns the list of functions you can use in the formula string }
    procedure GetFunctionList(var StList:TStringList);
  published
    { Published-Deklarationen }
    property Formula:AnsiString read FFormula write SetFFormula; // set new formula and do some action
    property Parameters:TStrings read FParameters write SetFParameters;
    property Constants:TConstantsList read FAddedConstants write FAddedConstants;
    property OnValidateParameter:TValidateParameter read FOnValidateParameter write FOnValidateParameter; // if Parameters is not set, you can use this function to validate a parameter
    property OnGetParameterValue:TGetParameterValue read FOnGetParameterValue write FOnGetParameterValue;
  end;





{ procedure AddFunction

Name : The name of the function to be used in the formula string
NumberOfParameters : the number of parameters the function expects. A value of -1 means,
                     that the function accepts an unlimited number of parameters
Address : address of the procedure where the new function is implemented
StackUsage : how many FPU-Registers does the new function use, this is needed to optimize stack-usage
SaveStack : Should the FPU-Stack be saved before the new function is called ?

The new function has to be declared as stdcall like this :

function NewFunction ( const A : Array of Extended ) : Extended; stdcall;

If you do not use stdcall TFastFormula will crash without notice.

Address is @NewFunction.

StackUsage :

The FPU-Stack holds 8 registers. TFastFormula optimizes the formula so that memorycalls are minimized.
This means that there may be values in the stack that may not be deleted.
Let's say you implement a function that needs 5 FPU-registers. If there are more than 3 values on the
stack before your function is called, all values exceeding 8 will be deleted, which will probably leed
to an exception when TFastFormula needs these values. If TFastFormula knows that your function needs
5 FPU-Registers it will optimize the formula so that there are never more than three values on the
FPU-stack before your function is called.

SaveStack :

If you for some reason do not know how many FPU-registers your function needs, set the value to
3 or 4. This should work with most Delphi build in functions, that you wish to use in your new
function. If TFastFormula crashes due to stack-overflow you can set SAVESTACK to TRUE.
This will cause TFastFormula to clear the stack before your function is called. After your function
returns, TFastFormula will reload the values. This of course will result in memory-access and should
only be used in rare cases, due to speed-loss.

}


{ procedure AddInlineFunction

Functions added with AddInlineFunction are the fastes method of using user functions.
The code you write will be added directly to the code TFastFormula generates.
The disadvantage is that you have to provide the function in OpCodes.
To retrieve these opcodes either write the function using the Delphi Assembler and look
at the code Delphi generated using the CPU-Window in Delphi.
Or look at http://www.intel.com where they hold a PDF file describing all mnemonics and
accompaning opcodes.
If you have any questions feel free to ask me :

FKroeger@tkc.de

Name : The name of your new function to be used in formula strings.
NumberOfParameters : Number of Parameters the new function expects. A value of -1
                     means that the function accepts an unlimited number of parameters
StackUsage : Again, the number of FPU-registers the new function needs
ParametersInStack : If set to true the parameters will be in the FPU-stack,
                    If set to false, the parameters will be transfered over a memory block.
                    ParametersInStack := False =>
                    EBX => address to the memory block where the parameters are
                    EDX => number of parameters you will find there
CodeSize : Size of the inline code in Bytes
Code : points to an array of Byte where you store your code
RegisterUsage : Set RegisterUsage depending on the registers you use.
                If you use and change e.g.:

                EAX  : Set BIT 0
                EBX  : Set BIT 1  <= can only be used when parameters are in FPU-stack
                ECX  : Set BIT 2
                EDX  : Set BIT 3

                TFastFormula gives you access to three constants you might need.
                These are 2.0, 10.0 and 0.5.

                The register EDI points to a memory block where those constants are stored.

                to use  2.0 :   FLD tbyte ptr [edi]  = $DB $2F
                       10.0 :   FLD tbyte ptr [edi + $0a] = DB 6F $0A
                        0.5 :   FLD tbyte ptr [edi + $14] = $DB $6F $14

                If you need any of these constants set BIT 7 of RegisterUsage to 1.

                e.g.: Lets say you use EAX and any of the constants :

                RegisterUsage:=128 + 1;     // BIT 7 and BIT 0



For Example : This is the code TFastFormula uses for the MIN function :

// ebx points to the parameter array in memory

DB 2B           fld tbyte ptr [ebx]           // load first parameter
83 FA 01        cmp edx, 0x01                 // if only one parameter
74 1F           je @@MinEND                   // then end
             @@MinLOOP1:
83 C3 0A        add ebx, 0x0a                 // increase ebx by on parameter
DB 2B           fld tbyte ptr [ebx]           // load the next parameter
D8 D1           fcom                          // compare current and last parameter
9B              fwait                         // wait until FPU has finished comparing
DF E0           fstsw ax                      // store status word in AX
9E              sahf                          // store AX in EFLAGS
77 0A           ja @@DeleteST                 // if current parameter > last parameter delete current parameter
DD D9           fstp st(1)                    // else delete last parameter
4A              dec edx                       // decrease counter
83 FA 00        cmp edx, 0x00                 // Have we tested all parameters ?
75 EB           jne @@MinLOOP1                // No, so jmp to MinLOOP1
EB 08           jmp @@MinEND                  // Yes, so exit
             @@DeleteSt:
DD D8           fstp st                       // delete current parameter
4A              dec edx                       // decrease counter
83 FA 00        cmp edx, 0x00                 // Have we tested all parameters ?
75 E1           jne @@MinLOOP                 // No, so jmp to MinLOOP1
             @@MinEND:



To implement this function you would have to follow these steps :

const CodeArray : array[0..37] of Byte = ( $DB, $2B, $83, $FA, $01, $74, $1F, $83,
                                           $C3, $0A, $DB, $2B, $D8, $D1, $9B, $DF,
                                           $E0, $9E, $77, $0A, $DD, $D9, $4A, $83,
                                           $FA, $00, $75, $EB, $EB, $08, $DD, $D8,
                                           $4A, $83, $FA, $00, $75, $E1 );

FastFormula1.AddInlineFunction ( 'newMIN',
                                 -1,
                                 2,
                                 false,
                                 38,
                                 @CodeArray,
                                 1 );   // Register EAX is used, EBX and EDX were given by TFastFormula,
                                        // so there is no need to mention that we have changed them

